open TestDriver

(******************************************************************************)
(** Types from .mli ***********************************************************)
(******************************************************************************)

(* these types are copied from the .mli; see the "Specification" section below
   for the interesting code. *)

module type AddableCore = sig
  type t
  val (+)   : t -> t -> t
  val (~-)  : t -> t
  val zero : t
  val (=)  : t -> t -> bool
end

module type AddableExtension = sig
  type t
  val (-)    : t -> t -> t
  val (<>)   : t -> t -> bool
end

module type Addable = sig
  type t
  include AddableCore      with type t := t
  include AddableExtension with type t := t
end

module type MultipliableCore = sig
  type t
  include AddableCore with type t := t
  val ( * ) : t -> t -> t
  val one   : t
end

module type MultipliableExtension = sig
  include AddableExtension
  val from_int : int -> t
end

module type Multipliable = sig
  type t
  include MultipliableCore      with type t := t
  include MultipliableExtension with type t := t
end

module type OrderedCore = sig
  include MultipliableCore

  val isPositive : t -> bool
end

module type OrderedExtension = sig
  include MultipliableExtension

  include Set.OrderedType with type t := t
  include Map.OrderedType with type t := t

  val (<)  : t -> t -> bool
  val (>)  : t -> t -> bool
  val (<=) : t -> t -> bool
  val (>=) : t -> t -> bool
end

module type Ordered = sig
  include OrderedCore
  include OrderedExtension with type t := t
end

module type DividableCore = sig
  include MultipliableCore

  val inv : t -> t
end

module type DividableExtension = sig
  include MultipliableExtension

  val (/) : t -> t -> t
end

module type Dividable = sig
  include DividableCore
  include DividableExtension with type t := t
end

module type NumberCore = sig
  type t

  include DividableCore with type t := t
  include OrderedCore   with type t := t
end

module type NumberExtension = sig
  include DividableExtension
  include OrderedExtension   with type t := t
end

module type Number = sig
  include NumberCore
  include NumberExtension with type t := t
end

(******************************************************************************)
(** Specification test implementations ****************************************)
(******************************************************************************)

module AddableSpec (C : AddableCore) (E : Examples with type t := C.t) = struct
  open C

  let plus_is_commutative a b   = a + b = b + a

  let plus_is_associative a b c = (a + b) + c = a + (b + c)

  let minus_is_plus_inverse a   = a + (-a) = zero

  let zero_is_plus_identity a   = a + zero = a

  TEST_UNIT = check2 E.examples plus_is_commutative
  TEST_UNIT = check3 E.examples plus_is_associative
  TEST_UNIT = check1 E.examples minus_is_plus_inverse
  TEST_UNIT = check1 E.examples zero_is_plus_identity
end

module MultipliableSpec (C : MultipliableCore) (E : Examples with type t := C.t) = struct
  open C
  include AddableSpec(C)(E)

  let times_is_associative a b c        = (a * b) * c = a * (b * c)

  let times_is_commutative a b          = a * b = b * a

  let times_distributes_over_plus a b c = a * (b + c) = a * b + a * c

  let one_is_times_identity a           = a * one = a

  TEST_UNIT = check3 E.examples times_is_associative
  TEST_UNIT = check2 E.examples times_is_commutative
  TEST_UNIT = check3 E.examples times_distributes_over_plus
  TEST_UNIT = check1 E.examples one_is_times_identity
end

module OrderedSpec (C : OrderedCore) (E : Examples with type t := C.t) = struct
  open C

  include MultipliableSpec(C)(E)

  let sum_of_positives_is_positive a b =
    if isPositive a && isPositive b then isPositive (a + b) else true

  let product_of_positives_is_positive a b =
    if isPositive a && isPositive b then isPositive (a * b) else true

  let everything_is_either_positive_negative_or_zero a =
    (isPositive a) || (isPositive (-a)) || (a = zero)

  TEST_UNIT = check2 E.examples sum_of_positives_is_positive
  TEST_UNIT = check2 E.examples product_of_positives_is_positive
  TEST_UNIT = check1 E.examples everything_is_either_positive_negative_or_zero
end

module DividableSpec (C : DividableCore) (E : Examples with type t := C.t) = struct
  open C

  include MultipliableSpec(C)(E)

  let inv_is_times_inverse a = if a = zero then true else a * inv a = one

  TEST_UNIT = check1 E.examples inv_is_times_inverse
end

module NumberSpec (C : NumberCore) (E : Examples with type t := C.t) = struct
  include OrderedSpec(C)(E)
  include DividableSpec(C)(E)
end

(******************************************************************************)
(** Extension implementations                                                 *)
(******************************************************************************)

module ExtendAddable (C : AddableCore)
  : Addable with type t := C.t =
struct
  include C

  let (-)  a b = a + (-b)
  let (<>) a b = not (a = b)
end

module ExtendMultipliable (C : MultipliableCore)
  : Multipliable with type t := C.t =
struct
  include ExtendAddable(C)
  include C

  let rec from_int n =
    let open Pervasives in
    if n = 0 then zero
    else if n < 0 then C.(~-) (from_int (-n))
    else C.(+) C.one (from_int (n-1))
end

module ExtendOrdered (C : OrderedCore)
 : Ordered with type t := C.t =
struct
  include C
  include ExtendMultipliable(C)

  let (>)  a b = isPositive (a - b)
  let (<)  a b = isPositive (b - a)
  let (>=) a b = a > b || a = b
  let (<=) a b = a < b || a = b

  let compare a b =
    if a > b then 1
    else if a = b then 0
    else -1
end

module ExtendDividable (C : DividableCore)
  : Dividable with type t := C.t =
struct
  include C
  include ExtendMultipliable(C)

  let (/) a b = a * (inv b)
end

module ExtendNumber (C : NumberCore)
  : Number with type t := C.t =
struct
  include C
  include ExtendOrdered(C)
  include ExtendDividable(C)
end


